struct Settings
    horizon::Base.RefValue{Int}
    use_forecast_data::Bool
    use_parameters::Base.RefValue{Bool}
    warm_start::Base.RefValue{Bool}
    balance_slack_variables::Bool
    services_slack_variables::Bool
    initial_time::Base.RefValue{Dates.DateTime}
    PTDF::Union{Nothing, PSY.PTDF}
    optimizer::Union{Nothing, JuMP.MOI.OptimizerWithAttributes}
    optimizer_log_print::Bool
    constraint_duals::Vector{Symbol}
    system_to_file::Bool
    export_pwl_vars::Bool
    allow_fails::Bool
    ext::Dict{String, Any}
end

function Settings(
    sys;
    initial_time::Dates.DateTime = UNSET_INI_TIME,
    use_parameters::Bool = false,
    use_forecast_data::Bool = true,
    warm_start::Bool = true,
    balance_slack_variables::Bool = false,
    services_slack_variables::Bool = false,
    horizon::Int = UNSET_HORIZON,
    PTDF::Union{Nothing, PSY.PTDF} = nothing,
    optimizer::Union{Nothing, JuMP.MOI.OptimizerWithAttributes} = nothing,
    optimizer_log_print::Bool = false,
    constraint_duals::Vector{Symbol} = Vector{Symbol}(),
    system_to_file = true,
    export_pwl_vars = false,
    allow_fails = false,
    ext = Dict{String, Any}(),
)
    return Settings(
        Ref(horizon),
        use_forecast_data,
        Ref(use_parameters),
        Ref(warm_start),
        balance_slack_variables,
        services_slack_variables,
        Ref(initial_time),
        PTDF,
        optimizer,
        optimizer_log_print,
        constraint_duals,
        system_to_file,
        export_pwl_vars,
        allow_fails,
        ext,
    )
end

function copy_for_serialization(settings::Settings)
    vals = []
    for name in fieldnames(Settings)
        if name == :optimizer
            # Cannot guarantee that the optimizer can be serialized.
            val = nothing
        else
            val = getfield(settings, name)
        end

        push!(vals, val)
    end

    return deepcopy(Settings(vals...))
end

function restore_from_copy(
    settings::Settings;
    optimizer::Union{Nothing, JuMP.MOI.OptimizerWithAttributes},
)
    vals = Dict{Symbol, Any}()
    for name in fieldnames(Settings)
        if name == :optimizer
            vals[name] = optimizer
        elseif name == :ext
            continue
        else
            val = getfield(settings, name)
            vals[name] = isa(val, Base.RefValue) ? val[] : val
        end
    end

    return vals
end

get_horizon(settings::Settings) = settings.horizon[]
get_use_forecast_data(settings::Settings) = settings.use_forecast_data
get_use_parameters(settings::Settings) = settings.use_parameters[]
get_initial_time(settings::Settings)::Dates.DateTime = settings.initial_time[]
get_PTDF(settings::Settings) = settings.PTDF
get_optimizer(settings::Settings) = settings.optimizer
get_ext(settings::Settings) = settings.ext
get_warm_start(settings::Settings) = settings.warm_start[]
get_constraint_duals(settings::Settings) = settings.constraint_duals
get_balance_slack_variables(settings::Settings) = settings.balance_slack_variables
get_services_slack_variables(settings::Settings) = settings.services_slack_variables
get_system_to_file(settings::Settings) = settings.system_to_file
get_export_pwl_vars(settings::Settings) = settings.export_pwl_vars
get_allow_fails(settings::Settings) = settings.allow_fails
get_optimizer_log_print(settings::Settings) = settings.optimizer_log_print

function set_horizon!(settings::Settings, horizon::Int)
    settings.horizon[] = horizon
    return
end

function set_initial_time!(settings::Settings, initial_time::Dates.DateTime)
    settings.initial_time[] = initial_time
    return
end

function set_use_parameters!(settings::Settings, val::Bool)
    settings.use_parameters[] = val
    return
end

function set_warm_start!(settings::Settings, warm_start::Bool)
    settings.warm_start[] = warm_start
    return
end
